"""Annotation parsers."""
import json
from typing import Any, Dict, List

from zod.anno.lane import LaneAnnotation, parse_lane_annotation
from zod.anno.road_condition import RoadConditionAnnotation
from zod.constants import AnnotationProject

from .object import ObjectAnnotation
from .tsr.traffic_sign import TrafficSignAnnotation


def _read_annotation_file(annotation_file: str) -> List[Dict[str, Any]]:
    """Read an annotation file."""
    with open(annotation_file, "r") as file:
        return json.load(file)


def parse_object_detection_annotation(annotation_path: str) -> List[ObjectAnnotation]:
    """Parse the objects annotation from the annotation string."""
    annotation = _read_annotation_file(annotation_path)
    return [ObjectAnnotation.from_dict(obj) for obj in annotation]


def parse_traffic_sign_annotation(annotation_path: str) -> List[TrafficSignAnnotation]:
    """Parse the traffic sign annotation from the annotation string."""
    annotation = _read_annotation_file(annotation_path)
    return [TrafficSignAnnotation.from_dict(sign) for sign in annotation]


def parse_lane_markings_annotation(annotation_path: str) -> List[LaneAnnotation]:
    """Parse the lane markings annotation from the annotation string."""
    annotations = _read_annotation_file(annotation_path)
    return [parse_lane_annotation(marker) for marker in annotations]


def parse_ego_road_annotation(annotation_path: str) -> List[Dict]:
    """Parse the egoroad annotation from the annotation string."""
    annotations = _read_annotation_file(annotation_path)
    polygons = []
    for annotation in annotations:
        if "class" in annotation["properties"]:
            polygons.append(annotation["geometry"]["coordinates"])
        else:
            # TODO: what does it mean if we end up here?
            pass
    return polygons


def parse_road_condition_annotation(annotation_path) -> RoadConditionAnnotation:
    with open(annotation_path, "r") as file:
        annotation = json.load(file)
    return RoadConditionAnnotation.from_dict(annotation["properties"])


ANNOTATION_PARSERS = {
    AnnotationProject.LANE_MARKINGS: parse_lane_markings_annotation,
    AnnotationProject.EGO_ROAD: parse_ego_road_annotation,
    AnnotationProject.OBJECT_DETECTION: parse_object_detection_annotation,
    AnnotationProject.TRAFFIC_SIGNS: parse_traffic_sign_annotation,
    AnnotationProject.ROAD_CONDITION: parse_road_condition_annotation,
}



"""Functions for evaluating object detection models."""

import json
import os
from typing import Callable, Dict

import numpy as np

from zod.eval.detection._nuscenes_eval.common.data_classes import EvalBoxes
from zod.eval.detection._nuscenes_eval.common.utils import center_distance
from zod.eval.detection._nuscenes_eval.detection.algo import accumulate, calc_ap, calc_tp
from zod.eval.detection._nuscenes_eval.detection.data_classes import (
    DetectionBox,
    DetectionConfig,
    DetectionMetrics,
)
from zod.eval.detection.constants import EVALUATION_CLASSES

VALID_TP_METRICS = {"trans_err": "mATE", "scale_err": "mASE", "orient_err": "mAOE"}
PRECISION_RECALL_SAMPLING_POINTS = 101

NUSCENES_SUMMARY = """
NUSCENES METRIC SUMMARY:
distance function: {dist_func}
distance thresholds: {dist_thresh}
--------------------------------------------------
NDS:  {NDS:.4f}
mAP:  {mAP:.4f}
mATE: {mATE:.4f}
mASE: {mASE:.4f}
mAOE: {mAOE:.4f}"""

PER_CLASS_SUMMARY = """
{cls}:
    mAP:  {mAP:.4f}
    mATE: {mATE:.4f}
    mASE: {mASE:.4f}
    mAOE: {mAOE:.4f}"""

NUSCENES_DEFAULT_SETTINGS = {
    "class_range": {
        "Vehicle": 50,
        "VulnerableVehicle": 40,
        "Pedestrian": 30,
        "TrafficSign": 30,
        "TrafficSignal": 30,
    },
    "dist_fcn": "center_distance",
    "dist_ths": [0.5, 1.0, 2.0, 4.0],
    "dist_th_tp": 2.0,
    "min_recall": 0.1,
    "min_precision": 0.1,
    "max_boxes_per_sample": 500,
    "mean_ap_weight": 5,
}


def evaluate_nuscenes_style(
    gt_boxes: EvalBoxes,
    det_boxes: EvalBoxes,
    verbose: bool = False,
    output_path: str = None,
    verify_coordinate_system: bool = True,
) -> Dict[str, float]:
    """Perform nuscenes evaluation based on a number of groundtruths and detections."""
    if verify_coordinate_system:
        _check_coordinate_system(gt_boxes)

    detection_cfg = DetectionConfig(**NUSCENES_DEFAULT_SETTINGS)
    detection_metrics = DetectionMetrics(detection_cfg)

    class_ranges = {k: (0, v) for k, v in NUSCENES_DEFAULT_SETTINGS["class_range"].items()}

    # filter according to the default nuscenes settings
    gt_boxes = _filter_eval_boxes_on_ranges(gt_boxes, class_ranges)
    det_boxes = _filter_eval_boxes_on_ranges(det_boxes, class_ranges)

    metrics = {
        dist_th: _nuscenes_evaluate(gt_boxes, det_boxes, dist_th=dist_th)
        for dist_th in detection_cfg.dist_ths
    }

    evaluated_clses = set(metrics[detection_cfg.dist_ths[0]].keys())
    for zod_cls in evaluated_clses:
        # They evaluate the ap across all thresholds
        for dist_th in detection_cfg.dist_ths:
            detection_metrics.add_label_ap(
                detection_name=zod_cls,
                dist_th=dist_th,
                ap=metrics[dist_th][zod_cls]["ap"],
            )

        # They evaluate the tp across only one threshold
        for metric in VALID_TP_METRICS:
            detection_metrics.add_label_tp(
                zod_cls, metric, metrics[detection_cfg.dist_th_tp][zod_cls][metric]
            )

    if verbose:
        _print_nuscenes_metrics(detection_metrics)

    serialized = _serialize(detection_metrics)

    if output_path:
        with open(os.path.join(output_path, "nuscenes_evaluation_metrics.json"), "w") as f:
            json.dump(serialized, f)

    return serialized


def _check_coordinate_system(gt_boxes: EvalBoxes):
    """Use heuristics to check if the boxes were provided in the ego coordinate system."""
    # The key assumption is x axis has the longest range in the ego coordinate system
    # whereas for camera and lidar the z and y axis are the longest range respectively
    centers = np.array([box.translation for box in gt_boxes.all])
    center_means = np.mean(centers, axis=0)
    max_avgdist_axis = np.argmax(center_means)
    error_msg = (
        "Looks like the boxes were provided in the {sensor} coordinate system. "
        "Please convert them to the ego coordinate system. If you know what you are doing, "
        " you can disable this check by setting `verify_coordinate_system=False`."
    )
    desired_axis = 0  # EGO
    if max_avgdist_axis == 0 and desired_axis != 0:
        raise ValueError(error_msg.format(sensor="ego"))
    elif max_avgdist_axis == 1 and desired_axis != 1:
        raise ValueError(error_msg.format(sensor="lidar"))
    elif max_avgdist_axis == 2 and desired_axis != 2:
        raise ValueError(error_msg.format(sensor="camera"))


def _nuscenes_evaluate(
    gt_boxes: EvalBoxes,
    det_boxes: EvalBoxes,
    dist_fcn: Callable = center_distance,
    dist_th: float = 1,
    min_precision: float = 0.1,
    min_recall: float = 0.1,
) -> Dict[str, float]:
    """Perform nuscenes evaluation based on a number of groundtruths and detections."""
    metrics = {}

    for cls in EVALUATION_CLASSES:
        # ensure that we have samples for this class in the gt data
        n_samples = sum(1 for box in gt_boxes.all if box.detection_name == cls)
        if n_samples == 0:
            continue
        md = accumulate(
            gt_boxes,
            det_boxes,
            cls,
            dist_fcn=dist_fcn,
            dist_th=dist_th,
        )

        metrics[cls] = {
            metric: calc_tp(md, min_recall=0.1, metric_name=metric) for metric in VALID_TP_METRICS
        }
        metrics[cls]["ap"] = calc_ap(md, min_recall=min_recall, min_precision=min_precision)
    return metrics


def _filter_eval_boxes_on_ranges(
    boxes: EvalBoxes, class_ranges: Dict[str, int], verbose: bool = False
) -> EvalBoxes:
    """Filter out boxes that are outside of the range of the classes."""
    filtered_boxes = EvalBoxes()

    def _filter(box: DetectionBox):
        return (
            box.detection_name in class_ranges
            and class_ranges[box.detection_name][0]
            < box.ego_dist
            <= class_ranges[box.detection_name][1]
        )

    for frame_id in boxes.sample_tokens:
        filtered_boxes.add_boxes(frame_id, [box for box in boxes[frame_id] if _filter(box)])
    if verbose:
        # print the number of removed boxes per class
        for cls in class_ranges:
            n_originally = sum(1 for box in boxes.all if box.detection_name == cls)
            n_filtered = sum(1 for box in filtered_boxes.all if box.detection_name == cls)
            n_removed = n_originally - n_filtered
            if n_removed > 0:
                print(f"{cls}: from {n_originally} -> {n_filtered} ({n_removed} removed)")

    return filtered_boxes


def _print_nuscenes_metrics(metrics: DetectionMetrics):
    print(
        NUSCENES_SUMMARY.format(
            dist_func=metrics.cfg.dist_fcn,
            dist_thresh=metrics.cfg.dist_ths,
            NDS=metrics.nd_score,
            mAP=metrics.mean_ap,
            mATE=metrics.tp_errors["trans_err"],
            mASE=metrics.tp_errors["scale_err"],
            mAOE=metrics.tp_errors["orient_err"],
        )
    )
    for cls in metrics.mean_dist_aps:
        print(
            PER_CLASS_SUMMARY.format(
                cls=cls,
                mAP=metrics.mean_dist_aps[cls],
                mATE=metrics._label_tp_errors[cls]["trans_err"],
                mASE=metrics._label_tp_errors[cls]["scale_err"],
                mAOE=metrics._label_tp_errors[cls]["orient_err"],
            )
        )


def _serialize(detection_metrics: DetectionMetrics) -> Dict[str, float]:
    # Only serialize the classes that were evaluated (had GT)
    classes = list(detection_metrics.mean_dist_aps.keys())
    tp_metrics = {
        name: detection_metrics.tp_errors[metric] for metric, name in VALID_TP_METRICS.items()
    }
    class_aps = {f"{cls}/mAP": detection_metrics.mean_dist_aps[cls] for cls in classes}
    class_tps = {
        f"{cls}/{name}": detection_metrics._label_tp_errors[cls][metric]
        for cls in classes
        for metric, name in VALID_TP_METRICS.items()
    }
    return {
        "NDS": detection_metrics.nd_score,
        "mAP": detection_metrics.mean_ap,
        **tp_metrics,
        **class_aps,
        **class_tps,
    }



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict

from zod.anno.tsr.class_map import get_class_idx
from zod.constants import Camera
from zod.data_classes.box import Box2D


@dataclass
class TrafficSignAnnotation:
    """Traffic sign dataclass."""

    bounding_box: Box2D
    traffic_sign_class: str
    traffic_sign_idx: int
    occlusion_ratio: str
    annotation_uuid: str
    electronic_sign: bool
    uuid: str
    unclear: bool

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> TrafficSignAnnotation:
        """Create a TrafficSignAnnotation from a dict."""
        return cls(
            unclear=data["properties"]["unclear"],
            bounding_box=Box2D.from_points(data["geometry"]["coordinates"], Camera.FRONT),
            traffic_sign_class=data["properties"]["class"],
            traffic_sign_idx=get_class_idx(data["properties"]["class"]),
            occlusion_ratio=data["properties"]["occlusion_ratio"],
            annotation_uuid=data["properties"]["annotation_uuid"],
            electronic_sign=data["properties"]["is_electronic"],
            uuid=data["properties"]["annotation_uuid"],
        )



"""ZOD Frames module."""
import json
import os.path as osp
from functools import partial
from itertools import repeat
from pathlib import Path
from typing import Dict, List, Set, Tuple, Union

from tqdm.contrib.concurrent import process_map

from zod.constants import FRAMES, FULL, TRAIN, TRAINVAL_FILES, VAL, VERSIONS
from zod.data_classes.frame import ZodFrame
from zod.data_classes.info import Information
from zod.utils.utils import zfill_id


def _create_frame(frame: dict, dataset_root: str) -> Information:
    info = Information.from_dict(frame)
    info.convert_paths_to_absolute(dataset_root)
    return info


class ZodFrames(object):
    """ZOD Frames."""

    def __init__(self, dataset_root: Union[Path, str], version: str, mp: bool = True):
        self._dataset_root = dataset_root
        self._version = version
        self._mp = mp
        assert version in VERSIONS, f"Unknown version: {version}, must be one of: {VERSIONS}"
        self._train_frames, self._val_frames = self._load_frames()
        self._frames: Dict[str, Information] = {**self._train_frames, **self._val_frames}

    def __len__(self) -> int:
        return len(self._frames)

    def __getitem__(self, frame_id: Union[int, str, slice]) -> ZodFrame:
        """Get frame by id, which is a 6-digit zero-padded number. Ex: '000001'."""
        if isinstance(frame_id, slice):
            return [self.__getitem__(i) for i in range(*frame_id.indices(len(self)))]
        frame_id = zfill_id(frame_id)
        return ZodFrame(self._frames[frame_id])

    def __iter__(self):
        for frame_id in self._frames:
            yield self.__getitem__(frame_id)

    def _load_frames(self) -> Tuple[Dict[str, Information], Dict[str, Information]]:
        """Load frames for the given version."""
        filename = TRAINVAL_FILES[FRAMES][self._version]
        with open(osp.join(self._dataset_root, filename), "r") as f:
            all_ids = json.load(f)

        func = partial(_create_frame, dataset_root=self._dataset_root)
        if self._mp and self._version == FULL:
            train_frames = process_map(
                func,
                all_ids[TRAIN],
                desc="Loading train frames",
                chunksize=50 if self._version == FULL else 1,
            )
            val_frames = process_map(
                func,
                all_ids[VAL],
                desc="Loading val frames",
                chunksize=50 if self._version == FULL else 1,
            )
            train_frames = {frame.id: frame for frame in train_frames}
            val_frames = {frame.id: frame for frame in val_frames}
        else:
            train_frames = {frame.id: frame for frame in map(func, all_ids[TRAIN])}
            val_frames = {frame.id: frame for frame in map(func, all_ids[VAL])}

        return train_frames, val_frames

    def get_split(self, split: str) -> List[str]:
        """Get split by name (e.g. train / val)."""
        if split == TRAIN:
            return list(self._train_frames.keys())
        elif split == VAL:
            return list(self._val_frames.keys())
        else:
            raise ValueError(f"Unknown split: {split}, should be {TRAIN} or {VAL}")

    def get_all_frame_infos(self) -> Dict[str, Information]:
        """Get all frames."""
        return self._frames

    def get_all_ids(self) -> Set[str]:
        """Get all frame ids."""
        return set(self._frames.keys())
